#27/02/2025
#EOS parent class
import sympy as sp
from alpha_0 import Alpha0

from symbols import *

class EOS:

    def __init__(self, name:str, fluid, cp0_object, AlphaRClass, **kwargs):
        self.fluid = fluid
        
        self.alpha_0 = Alpha0(self.fluid, cp0_object)
        self.alpha_r = AlphaRClass.for_purefluid(self.fluid, **kwargs)

        self.alpha = (self.alpha_0.alpha_0 + self.alpha_r.alpha_r)

        self.name = name

        self.pressure_equation = sp.lambdify((rho, P, T), P- self.pressure)

    @property
    def pressure(self):
        pressure_expression = rho*self.fluid.R*T*(1+ rho*sp.diff(self.alpha_r.alpha_r, rho))
        return pressure_expression

    @property
    def compressibility(self):
        compressibility_expression = (1+ rho*sp.diff(self.alpha_r.alpha_r, rho))
        return compressibility_expression

    @property
    def specific_heat(self):
        alpha_0 = self.alpha_0.alpha_0.subs([(T, 1/T_inv)])
        alpha_r = self.alpha_r.alpha_r.subs([(T, 1/T_inv)])

        cp_expression = self.fluid.R*(-T_inv**2*(sp.diff(alpha_0, T_inv, T_inv) + sp.diff(alpha_r, T_inv, T_inv)) + \
                        (1+rho*sp.diff(alpha_r, rho) - rho*T_inv*sp.diff(alpha_r, rho, T_inv))**2/ \
                        (1+2*rho*sp.diff(alpha_r, rho) +rho**2*sp.diff(alpha_r,rho, rho)))
        cp_expression = cp_expression.subs([(T_inv, 1/T)])

        return cp_expression

    @property
    def joule_thomson_coefficient(self):
        alpha_0 = self.alpha_0.alpha_0.subs([(T, 1/T_inv)])
        alpha_r = self.alpha_r.alpha_r.subs([(T, 1/T_inv)])

        muJT_expression = 1/(self.fluid.R*rho)*-(rho*sp.diff(alpha_r, rho) +rho**2*sp.diff(alpha_r, rho, rho)+rho*T_inv*sp.diff(alpha_r, rho, T_inv))/ \
                            ((1+rho*sp.diff(alpha_r,rho) -rho*T_inv*sp.diff(alpha_r, rho, T_inv))-T_inv**2*(sp.diff(alpha_0, T_inv, T_inv)+sp.diff(alpha_r,T_inv,T_inv))*\
                                (1+2*rho*sp.diff(alpha_r, rho)+rho**2*sp.diff(alpha_r, rho, rho)))

        muJT_expression = muJT_expression.subs([(T_inv, 1/T)])

        return muJT_expression

    @property
    def gibbs_free_energy(self):
        alpha_0 = self.alpha_0.alpha_0.subs([(T, 1/T_inv)])
        alpha_r = self.alpha_r.alpha_r.subs([(T, 1/T_inv)])

        g_expression = self.fluid.R*T*(1+ alpha_0 + alpha_r + rho*sp.diff(alpha_r, rho))

        g_expression = g_expression.subs([(T_inv, 1/T)])

        return g_expression

    #see log page 163
    @property
    def fugacity_coefficient(self):

        phi_expression = sp.exp(self.alpha_r.alpha_r + rho*sp.diff(self.alpha_r.alpha_r, rho) - sp.log(1+rho*sp.diff(self.alpha_r.alpha_r, rho)))

        return phi_expression  #fugacity coefficient phi as a function of density and temperature
    



#06/05/2025

#sympy symbols
from symbols import *

# cubic non dimensional residual Helmholts energy class
# necessary to create any cubic EOS for a pure fluid
class AlphaRCubic():
    """
    Class for calculating the residual Helmholtz energy of cubic equations of state.

    This class provides methods to compute the residual Helmholtz energy as a function of temperature and density.
    The expression for the Helmholtz energy is created by substituting the symbols for values included in the input dictionary.

    Attributes:
        expression (sp.core.add.Add): The expression for the residual Helmholtz energy.
        a (float): The parameter 'a' related to the equation of state.
        b (float): The parameter 'b' related to the equation of state.
        alpha_r (sp.core.add.Add): The evaluated residual Helmholtz energy after substituting 'a' and 'b' into the expression.

    Methods:
        for_purefluid(cls, fluid, alpha_r_expr, a_c_expr, alpha_T_expr, b_expr):
            Class method to create an instance for a pure fluid using the provided expressions and fluid properties.

        for_mixture(cls, mix, mixture_rule, cubic_alpha_r_list, alpha_r_expr, **kwargs):
            Class method to create an instance for a mixture using the provided mixture rule and expressions.
    """

    def __init__(self, alpha_r_expr: sp.core.add.Add, a_value,  b_value):
        self.expression = alpha_r_expr

        self.a = a_value
        self.b = b_value

        self.a_function = sp.lambdify((T, z1, z2), self.a)
        self.b_function = sp.lambdify((z1, z2), self.b)

        self.alpha_r = self.expression.subs([(a, self.a), (b, self.b)])

    #for a pure fluid
    @classmethod
    def for_purefluid(cls, fluid, alpha_r_expr: sp.core.add.Add, a_c_expr: sp.core.mul.Mul, alpha_T_expr: sp.core.power.Pow, b_expr: sp.core.mul.Mul):
        a_c_value = a_c_expr.subs([ (T_c, fluid.T_c), (P_c, fluid.P_c)])
        alpha_T_value = alpha_T_expr.subs([(omega, fluid.omega), (T_c, fluid.T_c)])

        a_value = a_c_value*alpha_T_value

        b_value = b_expr.subs([(T_c, fluid.T_c), (P_c, fluid.P_c)])

        return cls(alpha_r_expr, a_value, b_value)




    





#19/08/2025

from symbols import *

class AlphaRHelmholtz():
    
    def __init__(self, alpha_r_expr: sp.core.add.Add ):
        self.expression = alpha_r_expr  
        
        self.alpha_r = alpha_r_expr
        

    # for a pure fluid (fluid is just a placeholder argument)
    @classmethod
    def for_purefluid(cls, fluid, alpha_r_expr: sp.core.add.Add):
        # sub delta and tau for their pure fluid values using density and temperature
        return cls(alpha_r_expr)



import sympy as sp

# constants:
# universal gas constant
R = 8.31446261815324

# boltzmann constant in si
k_b = 1.380649e-23              # [J/K] -> [m2 kg/(s2 K)]

# boltzmann constant in Gaussian units
k_b_Gaussian = 1.380649e-16     # [erg/K]

# Avogadros number
N_av = 6.022140857*1e23         # Avogadro's number

# pi, in float
pi = float(sp.pi)

# symbols:
# sympy symbols used throuout all files
# they are explicitly here so its easier to keep all files consistent

#symbols for pressure, temperature, density, specific volume, compressibility factor
P, T, rho, v, Z = sp.symbols("P T rho v Z", positive=True)

#symbols for cubic equations of state
a_c, a, b = sp.symbols("a_c a b", positive = True)
alpha_T = sp.symbols("alpha_T")

#symbols for mixture composition, z: overall composition, x: liquid phase, y: vapour phase
z1, z2, x1, x2, y1, y2 = sp.symbols("z1 z2 x1 x2 y1 y2", positive = True)

# n: total number of moles, n1: moles of component 1, n2: moles of component 2
n, n1, n2 = sp.symbols("n n1 n2", positive = True)

# volume, mass, molar mass
V, m, M = sp.symbols("V m M", positive=True)

# molar density, molar volume
rho_m, v_m  = sp.symbols("rho_m v_m", positive=True)

# critical temperature, critical pressure, critical density
T_c, P_c, rho_c = sp.symbols("T_c P_c rho_c", positive=True)

# reduced variables
rho_r, T_r = sp.symbols("rho_r T_r", positive=True)
# acentric factor
omega = sp.symbols("omega")

# non dimensional density, temperature, and inverse temperature
delta, tau, T_inv = sp.symbols("delta tau T_inv", positive=True)


#27/02/2025
from dataclasses import dataclass

# class for storing pure fluid data
@dataclass
class PureFluid():
        """Class for creating pure fluids through their properties, all properties should be in SI units unless otherwise specified:

        name -> name of fluid
        formula -> molecular formula of fluid according to iupac
        M -> molar mass of fluid in [kg/m3] (SI)
        P_c -> pressure at critical point in [Pa]
        T_c -> temperature at critical point in [K]
        rho_c -> density at critical point in [kg/m3]
        P_t -> pressure at triple point in [Pa]
        T_t -> temperature at triple point [K]
        rho_t -> density at triple point in [kg/m3]
        omega -> accentric factor of pure fluid, see Wikipedia, also known as omega
        sigma -> Lennard-Jones characteristic length in [m], can be obtained from [0421] pages 779 onwards
        epsilon -> Lennard-Jones characteristic energy over boltzmann constant in [K], can be obtained from [0421] pages 779 onwards
        Sutherland_S -> Sutherland's constant used in the calculation of viscosity, see "Viscous Fluid Flow" for more information and numbers, in [K]
        Sutherland_T0 -> Sutherland's reference temperature used in the calculation of viscosity, in [K]
        Sutherland_mu0 -> Sutherland's reference viscosity used in the calculation of viscosity, in [Ns/m^2]
        dipole -> the dipole moment, measure of the polarity of a molecule, in [debyes]
        """        
        name: str
        formula: str
        M: float
        P_c: float
        T_c: float
        rho_c: float
        P_t: float
        T_t: float
        rho_t: float
        omega: float
        sigma: float
        epsilon: float
        Sutherland_S: float
        Sutherland_T0: float
        Sutherland_mu0: float
        dipole: float
        R: float = 8.31446261815324






#03/03/2025
import sympy as sp
from symbols import *   #sympy symbols
from .purefluid import PureFluid

from .storeroom.fluids import H2, NH3, CH4, N2, CO2, CH3OH , C2H5OH, C2H6, C8H18, C4H9OH, C7H16, H2O

class Cp0():

    def __init__(self, fluid_formula, cp0_expression, cp0_int_T_expresion=None, cp0_over_T_intT_T_expression=None):
        self.fluid_formula = fluid_formula
        self.cp0 = cp0_expression

        #simplify and expand
        if cp0_int_T_expresion == None or cp0_over_T_intT_T_expression==None:
            cp0_terms = sp.expand(sp.simplify(self.cp0)).args

            #integral of cp0 with respect to T
            integral_T = sum([sp.integrate(term, T) for term in cp0_terms])
            self.cp0_int_T = integral_T

            #integral of cp0 over T with respect to T
            integral_T = sum([sp.integrate(term/T, T) for term in cp0_terms])
            self.cp0_over_T_int_T = integral_T

        else:
            self.cp0_int_T = cp0_int_T_expresion
            self.cp0_over_T_int_T = cp0_over_T_intT_T_expression


    @classmethod
    def from_NIST(cls, fluid: type[PureFluid] , temperature_list: list[float, float, float],**kwargs):
        """
        ...
        Args:
            fluid (PureFluid object): object from PureFluid class

            temperature_list (list[float]): list of temperature ranges for NIST expression

            **A1 (float): coefficient A for the first temperature range
            **B1 (float): coefficient B for the first temperature range
            **C1 (float): coefficient C for the first temperature range
            **D1 (float): coefficient D for the first temperature range
            **E1 (float): coefficient E for the first temperature range


            **A2 (float): coefficient A for the second temperature range
            **B2 (float): coefficient B for the second temperature range
            **C2 (float): coefficient C for the second temperature range
            **D2 (float): coefficient D for the second temperature range
            **E2 (float): coefficient E for the second temperature range

        ...
        """

        # sigmoid, to avoid usage of heaviside
        def sigmoid(x):
            return 1/(1+sp.exp(-x))

        #create the expressions from the NIST arguments
        T0, T1, T2 = temperature_list
        A, B, C, D, E, a, b, c, d, e = kwargs.values()

        t = T/1000

        #if the temperature lies beyond the accpetable bounds it will still use the equation
        cp0 = (A+B*t+C*t**2 + D*t**3 + E/t**2)*sigmoid(-T+T1) + (a+b*t+c*t**2 + d*t**3 + e/t**2)*sigmoid(T-T1)
        cp0_int_T = (A*t + B*t**2/2 + C*t**3/3 +D*t**4/4 - E/t)*1000*sigmoid(-T+T1) + (a*t + b*t**2/2 + c*t**3/3 +d*t**4/4 - e/t)*1000*sigmoid(T-T1)
        cp0_over_T_int_T = (A*sp.log(T) + B*t + C/2*t**2 + D/3*t**3 - E/2/t**2)*sigmoid(-T+T1) + (a*sp.log(T) + b*t +c/2*t**2 + d/3*t**3 - e/2/t**2)*sigmoid(T-T1)

        return cls(fluid.formula, cp0, cp0_int_T, cp0_over_T_int_T)

    @classmethod
    def from_0421(cls, fluid: type[PureFluid] , temperature_list: list[float, float],**kwargs):
        """
        ...
        Args:
            fluid (PureFluid object): object from PureFluid class

            temperature_list (list[float]): lower and upper temperature limit for [0421] expression

            **a0 (float): coefficient a0
            **a1 (float): coefficient a1*1e3
            **a2 (float): coefficient a2*1e5
            **a3 (float): coefficient a3*1e8
            **a4 (float): coefficient a4*1e11
        ...
        """
        a0, a1, a2, a3, a4 = kwargs.values()
        cp0 = (a0 + a1*1e-3*T + a2*1e-5*T**2 + a3*1e-8*T**3 + a4*1e-11*T**4)*fluid.R
        cp0_int_T = (a0*T + a1*1e-3*T**2/2 + a2*1e-5*T**3/3 + a3*1e-8*T**4/4 + a4*1e-11*T**5/5)*fluid.R
        cp0_over_T_int_T = (a0*sp.log(T) + a1*1e-3*T + a2*1e-5*T**2/2 + a3*1e-8*T**3/3 + a4*1e-11*T**4/4)*fluid.R

        return cls(fluid.formula, cp0, cp0_int_T, cp0_over_T_int_T)




